#!/usr/bin/env python3

import sys
import argparse
import numpy as np

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import get_filename, glob_files

translate_parser = argparse.ArgumentParser(
    description="move atoms along a vector",
    formatter_class=argparse.RawTextHelpFormatter
)

translate_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

translate_parser.add_argument(
    "-if", "--input-format",
    type=str,
    choices=read_types,
    default=None,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

target_specifier = translate_parser.add_argument_group("atoms to move")
target_specifier.add_argument(
    "-t", "--targets",
    type=str,
    default=None,
    required=False,
    dest="targets",
    metavar="atoms",
    help="move atoms with specified indices\n"
    "Default: whole structure"
)

target_specifier.add_argument(
    "-f", "--fragment",
    type=str,
    default=None,
    required=False,
    dest="fragments",
    metavar="atoms",
    help="move fragments containing specified atoms\n"
    "Default: whole structure"
)

translate_parser.add_argument(
    "-ct", "--center-targets",
    type=str,
    default=None,
    required=False,
    dest="targets",
    metavar="targets",
    help="target atoms for -com or -cent arguments\n" +
    "comma (,) and/or hyphen (-) separated list\n" +
    "hyphens denote a range of atoms\n" +
    "commas separate individual atoms or ranges\n" +
    "default: whole structure"
)

translate_mode = translate_parser.add_argument_group(
    "translation mode (default: move centroid to origin)"
)
trans_modes = translate_mode.add_mutually_exclusive_group(required=True)
trans_modes.add_argument(
    "-v", "--vector",
    type=float,
    nargs=3,
    default=None,
    required=False,
    dest="vector",
    metavar=("x", "y", "z"),
    help="translate in direction of this vector\n" +
    "vector is normalized when --distance/-d is used"
)

translate_parser.add_argument(
    "-d", "--distance",
    type=float,
    default=None,
    required=False,
    dest="distance",
    help="distance translated - only applies to --vector/-v"
)

trans_modes.add_argument(
    "-dest", "--destination",
    type=float,
    nargs=3,
    default=None,
    required=False,
    dest="dest",
    metavar=("x", "y", "z"),
    help="translate fragment to a point"
)

center_parser = translate_parser.add_argument_group("center (default: centroid)")
center_option = center_parser.add_mutually_exclusive_group(required=False)
center_option.add_argument(
    "-com", "--center-of-mass",
    action="store_const",
    const=True,
    default=None,
    required=False,
    dest="com",
    help="translate the center of mass of the targets to the destination"
)

center_option.add_argument(
    "-cent", "--centroid",
    action="store_const",
    const=True,
    default=None,
    required=False,
    dest="cent",
    help="translate the centroid of the targets to the destination"
)

translate_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

args = translate_parser.parse_args()

for f in glob_files(args.infile, parser=translate_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format, None))
        else:
            infile = FileReader(f)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format, f))
        else:
            infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile)

    # targets to move
    targets = []
    if args.targets is not None:
        targets.extend(geom.find(args.targets))
    
    if args.fragments is not None:
        for atom in geom.find(args.fragments):
            frag_atoms = geom.get_all_connected(atom)
            targets.extend([frag_atom for frag_atom in frag_atoms if frag_atom not in targets])
    
    if not targets:
        targets = None

    # start with com or centroid where it is
    if args.com or not args.cent:
        com = geom.COM(targets=targets, mass_weight=True)
        start = com

    if args.cent:
        cent = geom.COM(targets=targets, mass_weight=False)
        start = cent

    # find where we are moving com or centroid to
    if args.dest is not None:
        # destination was specified
        destination = np.array(args.dest)
    elif args.vector is not None:
        # direction was specified
        if args.distance is not None:
            # magnitute was specified
            v = np.array(args.vector)
            v /= np.linalg.norm(v)
            destination = start + args.distance * v
        else:
            destination = start + np.array(args.vector)
    else:
        # nothing was specified - move to origin
        destination = np.zeros(3)

    translate_vector = destination - start

    geom.coord_shift(translate_vector, targets=targets)

    if args.outfile:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        geom.write(
            append=True,
            outfile=outfile,
        )
    else:
        print(geom.write(outfile=False))



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.component import Component
from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.utils.utils import get_filename, glob_files

def get_matching_ligands(name):
    name_info = name
    coordinating_elements = None
    denticity = None
    if isinstance(name_info, str):
        if (
                "name:" not in name_info.lower() and
                "elements:" not in name_info.lower() and
                "denticity:" not in name_info.lower()
        ):
            name_info = "^%s$" % name_info
        else:
            lig_info = name.split(":")
            for i, info in enumerate(lig_info):
                if info.lower() == "name" and i+1 < len(lig_info):
                    name_info = name_info.replace("%s:%s" % (info, lig_info[i+1]), "")
                    name_info = lig_info[i+1]

                elif info.lower() == "elements" and i+1 < len(lig_info):
                    name_info = name_info.replace("%s:%s" % (info, lig_info[i+1]), "")
                    coordinating_elements = lig_info[i+1].split(",")

                elif info.lower() == "denticity" and i+1 < len(lig_info):
                    name_info = name_info.replace("%s:%s" % (info, lig_info[i+1]), "")
                    denticity = int(lig_info[i+1])

            if not name_info:
                name_info = None

    return Component.list(
        name_regex=name_info,
        coordinating_elements=coordinating_elements,
        denticity=denticity
    )


maplig_parser = argparse.ArgumentParser(
    description="replace a ligand on an organometallic system",
    formatter_class=argparse.RawTextHelpFormatter,
)

maplig_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

maplig_parser.add_argument(
    "-ls",
    "--list",
    nargs="?",
    const=None,
    default=False,
    required=False,
    dest="list_avail",
    metavar="elements:X[,Y...] | name:RegEx",
    help="list available ligands\n" +
    "elements:X[,Y] can be used to only list ligands that coordinate\n" +
    "with the specified elements - must match exactly\n" +
    "name:RegEx can be used to only list ligands with names matching\n" +
    "the supplied regular expression - matches are case-insensitive",
)

maplig_parser.add_argument(
    "-if",
    "--input-format",
    type=str,
    nargs=1,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin",
)

maplig_parser.add_argument(
    "-l",
    "--ligand",
    metavar="[n[,m...]]=ligand | ligand",
    type=str,
    default=None,
    required=False,
    dest="ligand",
    help="ligand used to replace the current one\n" +
    "n[,m...] are the 1-indexed positions of the coordinating atoms of the\n" +
    "ligand that is being replaced\n" +
    "if these indices are not provided, they will the guessed\n" +
    "elements:X[,Y] or name:RegEx can be used in place of ligand\n" +
    "to swap ligands matching these criteria (see --list option)",
)

maplig_parser.add_argument(
    "-c",
    "--center",
    required=False,
    default=None,
    dest="center",
    help="catalyst center the ligand is bonded to\nDefault: any transition metal"
)

maplig_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=False,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$LIGAND will be replaced with ligand name\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout",
)

args = maplig_parser.parse_args()

if args.list_avail is not False:
    s = ""
    for i, name in enumerate(sorted(get_matching_ligands(args.list_avail))):
        s += "%-35s" % name
        if (i + 1) % 3 == 0:
        # if (i + 1) % 1 == 0:
            s += "\n"

    print(s.strip())
    sys.exit(0)

for infile in glob_files(args.infile, parser=maplig_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format[0], infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format[0], infile))
        else:
            f = FileReader(("from stdin", "xyz", infile))

    cat = Geometry(f)
    
    if args.center:
        cat.detect_components(center=args.center)
    
    # TODO: change this if to a regex
    if '=' in args.ligand:
        key_atoms = args.ligand.split("=")[0]
        lig_names = [l for l in "=".join(args.ligand.split("=")[1:]).split(",")]

    else:
        lig_names = [l for l in "=".join(args.ligand.split("=")[1:]).split(",")]
        key_atoms = []

    for lig_names in [get_matching_ligands(lig_name) for lig_name in lig_names]:
        for lig_name in lig_names:
            ligands = [Component(lig_name)]
            cat_copy = cat.copy()

            if key_atoms != []:
                key = cat_copy.find(key_atoms)
            else:
                key = []

            original_ligands = [l for l in ligands]

            lig_keys = sum([len(ligand.key_atoms) for ligand in ligands])
            while len(key) > lig_keys:
                ligands.extend(l.copy() for l in original_ligands)
                lig_keys += sum([len(ligand.key_atoms) for ligand in original_ligands])

            j = 0
            while len(key) < sum([len(ligand.key_atoms) for ligand in ligands]):
                if j >= len(cat.components["ligand"]):
                    raise RuntimeError(
                        "new ligand appears to have a higher denticity than old ligands combined"
                    )

                key.extend(cat.components["ligand"][j].key_atoms)
                j += 1

            cat_copy.map_ligand(ligands, key)

            if args.outfile:
                if "$INFILE" in args.outfile:
                    outfile = args.outfile.replace("$INFILE", get_filename(infile))
                outfile = outfile.replace("$LIGAND", lig_name)
                cat_copy.write(append=False, outfile=outfile)
            else:
                s = cat_copy.write(outfile=False)
                print(s)



import os
import re
import subprocess
from time import sleep

from jinja2 import Environment, FileSystemLoader, Template

from AaronTools import addlogger
from AaronTools.const import AARONLIB

USER = os.getenv("USER")
QUEUE_TYPE = os.getenv("QUEUE_TYPE", "None").upper()


class JobControl:
    pass


@addlogger
class SubmitProcess:
    """class for submitting jobs to the queue
    attributes:

    name:       name of job and input file minus the extension
    exe:        type of input file (com, in, inp)
    directory:  directory the input file is in
    walltime:   allocated walltime in hours
    processors: allocated processors
    memory:     allocated memory in GB
    template:   template job file"""

    LOG = None

    def __init__(self, fname, walltime, processors, memory, template=None):
        """fname:   str     - path to input file (e.g. /home/CoolUser/CoolStuff/neat.com
        walltime:   int/str - walltime in hours
        processors: int/str - allocated processors
        memory:     int/str - allocated memory in GB
        template:   str     - path to template file; if template is None, will look for
                              psi4.job, orca.job, or gaussian.job (depending on
                              extension on fname)"""
        directory, filename = os.path.split(fname)
        self.name, exe = os.path.splitext(filename)
        self.exe = exe[1:]
        self.directory = os.path.abspath(directory)
        self.walltime = walltime
        self.processors = processors
        self.memory = memory
        self.template = template
        if not isinstance(template, Template):
            self.set_template(template)

    @staticmethod
    def unfinished_jobs_in_dir(directory, retry=True):
        """returns list(jobids (str)) of jobids in directory
        retry: bool - if there's an error while checking the queue, sleep 300s and try again"""
        if QUEUE_TYPE == "LSF":
            args = ["bjobs", "-l", "2"]
            proc = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = proc.communicate()
            if len(err) != 0 and retry:
                SubmitProcess.LOG.warning(
                    "error checking queue: %s\nsleeping 300s before trying again"
                    % err.decode("utf-8")
                )
                sleep(300)
                return SubmitProcess.unfinished_jobs_in_dir(directory, retry)

            else:
                out = out.decode("utf-8")
                out = out.replace("\s", "")
                out = out.replace("\r", "")
                out = out.replace("\n", "")

                jobs = re.findall("(Job<\d+>.*RUNLIMIT)", out)

                job_ids = []
                for job in jobs:
                    test = re.match("Job<(\d+)>\S+CWD<.+%s>" % directory, job)
                    if test:
                        job_ids.append(test.group(1))

                return job_ids

        elif QUEUE_TYPE == "PBS":
            args = ["qstat", "-fx"]
            proc = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = proc.communicate()
            if len(err) != 0 and retry:
                SubmitProcess.LOG.warning(
                    "error checking queue: %s\nsleeping 300s before trying again"
                    % err.decode("utf-8")
                )
                sleep(300)
                return SubmitProcess.unfinished_jobs_in_dir(directory, retry)

            else:
                out = out.decode("utf-8")
                out = out.replace("\n", "")
                out = out.replace("\r", "")

                jobs = re.findall("<Job>(.+?)<\/Job>", out)

                job_ids = []
                for job in jobs:
                    # Q - queued
                    # R - running
                    # S - suspended
                    test = re.match(
                        "<Job_Id>(\d+).+<job_state>[QRS].+PBS_O_WORKDIR=[^,<>]*%s"
                        % directory,
                        job,
                    )
                    if test:
                        job_ids.append(test.group(1))

                return job_ids

        elif QUEUE_TYPE == "SLURM":
            args = ["squeue", "-o", "%i#%Z#%t", "-u", USER]
            proc = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = proc.communicate()
            if len(err) != 0 and retry:
                SubmitProcess.LOG.warning(
                    "error checking queue: %s\nsleeping 300s before trying again"
                    % err.decode("utf-8")
                )
                sleep(300)
                return SubmitProcess.unfinished_jobs_in_dir(directory, retry)

            else:
                out = out.decode("utf-8")
                job_ids = []
                for job in out.splitlines():
                    jobid, job_path, job_status = job.split("#")
                    if directory.endswith(job_path) and job_status in [
                        "R",
                        "PD",
                    ]:
                        job_ids.append(jobid)

                return job_ids

        elif QUEUE_TYPE == "SGE":
            # for SGE, we first grab the job ids for the jobs the user is running
            # we then call qstat again to get the directory those jobs are in
            args = ["qstat", "-s", "pr", "-u", USER]
            proc = subprocess.Popen(
                args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            out, err = proc.communicate()
            if len(err) != 0 and retry:
                SubmitProcess.LOG.warning(
                    "error checking queue: %s\nsleeping 300s before trying again"
                    % err.decode("utf-8")
                )
                sleep(300)
                return SubmitProcess.unfinished_jobs_in_dir(directory, retry)

            else:
                out = out.decode("utf-8")
                jobs = re.findall("^\s*?(\w+)", out)
                # first line is a header
                jobs.pop(0)

                jlist = ",".join(jobs)

                args = ["qstat", "-j", jlist]
                proc = subprocess.Popen(
                    args, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                out, err = proc.communicate()
                if len(err) != 0 and retry:
                    SubmitProcess.LOG.warning(
                        "error checking queue: %s\nsleeping 300s before trying again"
                        % err.decode("utf-8")
                    )
                    sleep(300)
                    return SubmitProcess.unfinished_jobs_in_dir(
                        directory, retry
                    )

                else:
                    out = out.decode("utf-8")
                    job_ids = []
                    for line in out.splitlines():
                        job_number = re.search("job_number:\s+(\d+)", line)
                        workdir = re.search(
                            "sge_o_workdir:\s+[\S]+%s$" % directory, line
                        )
                        if job_number:
                            job = job_number.group(1)

                        if workdir:
                            job_ids.append(job)

                    return job_ids

    def submit(self, wait=False, quiet=True, **opts):
        """submit job to the queue
        wait: bool/int - do not leave the function until any job in the directory
                         finishes (polled every 5 minutes or 'wait' seconds)
        opts: dict() used to render template; keys are template variables (e.g. exec_memory)
              and values are the corresponding values
        """

        job_file = os.path.join(self.directory, self.name + ".job")

        opts["name"] = self.name
        opts["walltime"] = self.walltime
        opts["processors"] = self.processors
        opts["memory"] = self.memory

        tm = self.template.render(**opts)

        if not os.path.isdir(self.directory):
            os.mkdirs(self.directory)
        
        with open(job_file, "w") as f:
            f.write(tm)

        stdin = None
        if QUEUE_TYPE == "LSF":
            args = ["bsub"]
            stdin = open(job_file, "r")
        elif QUEUE_TYPE == "SLURM":
            args = ["sbatch", job_file]
        elif QUEUE_TYPE == "PBS":
            args = ["qsub", job_file]
        elif QUEUE_TYPE == "SGE":
            args = ["qsub", job_file]
        else:
            raise NotImplementedError(
                "%s queues not supported, only LSF, SLURM, PBS, and SGE"
                % QUEUE_TYPE
            )

        proc = subprocess.Popen(
            args,
            cwd=self.directory,
            stdin=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        self.submit_out, self.submit_err = proc.communicate()

        if len(self.submit_err) != 0:
            raise RuntimeError(
                "error with submitting job %s: %s"
                % (self.name, self.submit_err.decode("utf-8"))
            )

        if not quiet:
            print(self.submit_out.decode("utf-8").strip())

        if wait is not False:
            if wait is True:
                wait_time = 30
            else:
                wait_time = abs(wait)

            sleep(wait_time)
            while len(self.unfinished_jobs_in_dir(self.directory)) != 0:
                # print(self.unfinished_jobs_in_dir(self.directory))
                sleep(wait_time)

        return

    def set_template(self, filename):
        """
        sets job template to filename
        AARONLIB directories are searched
        """
        if len(filename.splitlines()) > 1:
            # filename is actually the contents of a template file
            self.template = Template(filename)
            return

        # default templates are loaded from Aaron_Lib
        environment = Environment(loader=FileSystemLoader(AARONLIB))
        if filename is None:
            if self.exe == "com" or self.exe == "gjf":
                filename = "Gaussian_template.txt"
            elif self.exe == "inp":
                filename = "ORCA_template.txt"
            elif self.exe == "in":
                filename = "Psi4_template.txt"

        self.template = environment.get_template(filename)



#!/usr/bin/env python3

import argparse
import sys

from AaronTools.comp_output import CompOutput
from AaronTools.fileIO import FileReader
from AaronTools.spectra import Frequency
from AaronTools.utils.utils import get_filename, glob_files

from matplotlib import rcParams
import matplotlib.pyplot as plt

import numpy as np

rcParams["savefig.dpi"] = 300


peak_types = ["pseudo-voigt", "gaussian", "lorentzian", "delta"]
plot_types = ["transmittance", "absorbance", "vcd", "raman"]
weight_types = ["electronic", "zero-point", "enthalpy", "free", "quasi-rrho", "quasi-harmonic"]

def peak_type(x):
    out = [y for y in peak_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "peak type must be one of: %s" % ", ".join(
            peak_types
        )
    )

def plot_type(x):
    out = [y for y in plot_types if y.startswith(x)]
    if out:
        return out[0]
    raise TypeError(
        "plot type must be one of: %s" % ", ".join(
            plot_types
        )
    )

def weight_type(x):
    out = [y for y in weight_types if y.startswith(x)]
    if len(out) == 1:
        return out[0]
    raise TypeError(
        "weight type must be one of: %s" % ", ".join(
            weight_types
        )
    )


ir_parser = argparse.ArgumentParser(
    description="plot a Boltzmann-averaged IR spectrum",
    formatter_class=argparse.RawTextHelpFormatter
)

ir_parser.add_argument(
    "infiles", metavar="files",
    type=str,
    nargs="+",
    help="frequency job output file(s)"
)

ir_parser.add_argument(
    "-o", "--output",
    type=str,
    default=None,
    dest="outfile",
    help="output destination\n"
    "if the file extension is .csv, a CSV file will be written\n"
    "Default: show plot",
)

ir_parser.add_argument(
    "-t", "--plot-type",
    type=plot_type,
    choices=plot_types,
    default="transmittance",
    dest="plot_type",
    help="type of plot\nDefault: transmittance",
)


# TODO: figure out more anharmonic options
# anharmonic_options = ir_parser.add_argument_group("anharmonic options")
ir_parser.add_argument(
    "-na", "--harmonic",
    action="store_false",
    default=True,
    dest="anharmonic",
    help="force to use harmonic frequencies when anharmonic data is in the file",
)


peak_options = ir_parser.add_argument_group("peak options")
peak_options.add_argument(
    "-p", "--peak-type",
    type=peak_type,
    choices=peak_types,
    default="pseudo-voigt",
    dest="peak_type",
    help="function for peaks\nDefault: pseudo-voigt",
)

peak_options.add_argument(
    "-m", "--voigt-mixing",
    type=float,
    default=0.5,
    dest="voigt_mixing",
    help="fraction of pseudo-Voigt that is Gaussian\nDefault: 0.5",
)

peak_options.add_argument(
    "-fwhm", "--full-width-half-max",
    type=float,
    default=15.0,
    dest="fwhm",
    help="full width at half max. of peaks\nDefault: 15 cm^-1",
)

ir_parser.add_argument(
    "-s", "--point-spacing",
    default=None,
    type=float,
    dest="point_spacing",
    help="spacing between each x value\n"
    "Default: a non-uniform spacing that is more dense near peaks",
)


scale_options = ir_parser.add_argument_group("scale frequencies")
scale_options.add_argument(
    "-l", "--linear-scale",
    type=float,
    default=0.0,
    dest="linear_scale",
    help="subtract linear_scale * frequency from each mode (i.e. this is 1 - λ)\n"
    "Default: 0 (no scaling)",
)

scale_options.add_argument(
    "-q", "--quadratic-scale",
    type=float,
    default=0.0,
    dest="quadratic_scale",
    help="subtract quadratic_scale * frequency^2 from each mode\n"
    "Default: 0 (no scaling)",
)

ir_parser.add_argument(
    "-nr", "--no-reverse",
    action="store_false",
    default=True,
    dest="reverse_x",
    help="do not reverse x-axis",
)

center_centric = ir_parser.add_argument_group("x-centered interruptions")
center_centric.add_argument(
    "-sc", "--section-centers",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="centers",
    default=None,
    help="split plot into sections with a section centered on each of the specified values\n"
    "values should be separated by commas"
)

center_centric.add_argument(
    "-sw", "--section-widths",
    type=lambda x: [float(v) for v in x.split(",")],
    dest="widths",
    default=None,
    help="width of each section specified by -sc/--section-centers\n"
    "should be separated by commas, with one for each section"
)

minmax_centric = ir_parser.add_argument_group("x-range interruptions")
minmax_centric.add_argument(
    "-r", "--ranges",
    type=lambda x: [[float(v) for v in r.split("-")] for r in x.split(",")],
    dest="ranges",
    default=None,
    help="split plot into sections (e.g. 0-1900,2900-3300)"
)

ir_parser.add_argument(
    "-fw", "--figure-width",
    type=float,
    dest="fig_width",
    help="width of figure in inches"
)

ir_parser.add_argument(
    "-fh", "--figure-height",
    type=float,
    dest="fig_height",
    help="height of figure in inches"
)

ir_parser.add_argument(
    "-csv", "--experimental-csv",
    type=str,
    nargs="+",
    dest="exp_data",
    help="CSV file containing observed spectrum data, which will be plotted on top\n"
    "frequency job files should not come directly after this flag"
)

energy_options = ir_parser.add_argument_group("energy weighting")
energy_options.add_argument(
    "-w", "--weighting-energy",
    type=weight_type,
    dest="weighting",
    default="quasi-rrho",
    choices=weight_types,
    help="type of energy to use for Boltzmann weighting\n"
    "Default: quasi-rrho",
)

energy_options.add_argument(
    "-sp", "--single-point-files",
    type=str,
    nargs="+",
    default=None,
    required=False,
    dest="sp_files",
    help="single point energies to use for thermochem\n"
    "Default: energies from INFILES"
)

energy_options.add_argument(
    "-temp", "--temperature",
    type=float,
    dest="temperature",
    default=298.15,
    help="temperature (K) to use for weighting\n"
    "Default: 298.15",
)

energy_options.add_argument(
    "-w0", "--frequency-cutoff",
    type=float,
    dest="w0",
    default=100,
    help="cutoff frequency for quasi free energy corrections (1/cm)\n" +
    "Default: 100 cm^-1",
)

ir_parser.add_argument(
    "-rx", "--rotate-x-ticks",
    action="store_true",
    dest="rotate_x_ticks",
    default=False,
    help="rotate x-axis tick labels by 45 degrees"
)

args = ir_parser.parse_args()

if bool(args.centers) != bool(args.widths):
    sys.stderr.write(
        "both -sw/--section-widths and -sc/--section-centers must be specified"
    )
    sys.exit(2)

if args.ranges and bool(args.ranges) == bool(args.widths):
    sys.stderr.write(
        "cannot use -r/--ranges with -sw/--section-widths"
    )
    sys.exit(2)
    
centers = args.centers
widths = args.widths
if args.ranges:
    centers = []
    widths = []
    for (xmin, xmax) in args.ranges:
        centers.append((xmin + xmax) / 2)
        widths.append(abs(xmax - xmin))

exp_data = None
if args.exp_data:
    exp_data = []
    for f in args.exp_data:
        data = np.loadtxt(f, delimiter=",")
        for i in range(1, data.shape[1]):
            exp_data.append((data[:,0], data[:,i], None))

compouts = []
for f in glob_files(args.infiles, parser=ir_parser):
    fr = FileReader(f, just_geom=False)
    co = CompOutput(fr)
    compouts.append(co)

sp_cos = compouts
if args.sp_files:
    sp_cos = []
    for f in glob_files(args.sp_files, parser=ir_parser):
        fr = FileReader(f, just_geom=False)
        co = CompOutput(fr)
        sp_cos.append(co)

if args.weighting == "electronic":
    weighting = CompOutput.ELECTRONIC_ENERGY
elif args.weighting == "zero-point":
    weighting = CompOutput.ZEROPOINT_ENERGY
elif args.weighting == "enthalpy":
    weighting = CompOutput.RRHO_ENTHALPY
elif args.weighting == "free":
    weighting = CompOutput.RRHO
elif args.weighting == "quasi-rrho":
    weighting = CompOutput.QUASI_RRHO
elif args.weighting == "quasi-harmonic":
    weighting = CompOutput.QUASI_HARMONIC

for i, (freq, sp) in enumerate(zip(compouts, sp_cos)):
    rmsd = freq.geometry.RMSD(sp.geometry, sort=True)
    if rmsd > 1e-2:
        print(
            "single point energy structure might not match frequency file:\n"
            "%s %s RMSD = %.2f" % (sp.geometry.name,  freq.geometry.name, rmsd)
        )
    for freq2 in compouts[:i]:
        rmsd = freq.geometry.RMSD(freq2.geometry, sort=True)
        if rmsd < 1e-2:
            print(
                "two frequency files appear to be identical:\n"
                "%s %s RMSD = %.2f" % (freq2.geometry.name,  freq.geometry.name, rmsd)
            )

weights = CompOutput.boltzmann_weights(
    compouts,
    nrg_cos=sp_cos,
    temperature=args.temperature,
    weighting=weighting,
    v0=args.w0,
)

data_attr = "data"
if all(co.frequency.anharm_data for co in compouts) and args.anharmonic:
    data_attr = "anharm_data"


mixed_freq = Frequency.get_mixed_signals(
    [co.frequency for co in compouts],
    weights=weights,
    data_attr=data_attr,
)

if not args.outfile or not args.outfile.lower().endswith("csv"):
    fig = plt.gcf()
    fig.clear()
    
    mixed_freq.plot_ir(
        fig,
        centers=centers,
        widths=widths,
        plot_type=args.plot_type,
        peak_type=args.peak_type,
        reverse_x=args.reverse_x,
        fwhm=args.fwhm,
        point_spacing=args.point_spacing,
        voigt_mixing=args.voigt_mixing,
        linear_scale=args.linear_scale,
        quadratic_scale=args.quadratic_scale,
        exp_data=exp_data,
        anharmonic=mixed_freq.anharm_data and args.anharmonic,
        rotate_x_ticks=args.rotate_x_ticks,
    )
    
    if args.fig_width:
        fig.set_figwidth(args.fig_width)
    
    if args.fig_height:
        fig.set_figheight(args.fig_height)
    
    if args.outfile:
        plt.savefig(args.outfile, dpi=300)

    else:
        plt.show()

else:
    intensity_attr = "intensity"
    if args.plot_type.lower() == "vcd":
        intensity_attr = "rotation"
    if args.plot_type.lower() == "raman":
        intensity_attr = "raman_activity"

    funcs, x_positions, intensities = mixed_freq.get_spectrum_functions(
        fwhm=args.fwhm,
        peak_type=args.peak_type,
        voigt_mixing=args.voigt_mixing,
        linear_scale=args.linear_scale,
        quadratic_scale=args.quadratic_scale,
        data_attr="anharm_data" if mixed_freq.anharm_data and args.anharmonic else "data",
        intensity_attr=intensity_attr,
    )

    x_values, y_values, _ = mixed_freq.get_plot_data(
        funcs,
        x_positions,
        point_spacing=args.point_spacing,
        transmittance=args.plot_type == "transmittance",
        peak_type=args.peak_type,
        fwhm=args.fwhm,
    )

    if args.plot_type.lower() == "transmittance":
        y_label = "Transmittance (%)"
    elif args.plot_type.lower() == "absorbance":
        y_label = "Absorbance (arb.)"
    elif args.plot_type.lower() == "vcd":
        y_label = "delta_Absorbance (arb.)"
    elif args.plot_type.lower() == "raman":
        y_label = "Activity (arb.)"


    with open(args.outfile, "w") as f:
        s = ",".join(["frequency (cm^-1)", y_label])
        s += "\n"
        for x, y in zip(x_values, y_values):
            s += ",".join(["%.4f" % z for z in [x, y]])
            s += "\n"
        f.write(s)




#!/usr/bin/env python3
import os
import unittest

from AaronTools.geometry import Geometry
from AaronTools.test import TestWithTimer, prefix, rmsd_tol
from AaronTools.finders import *


class TestFinder(TestWithTimer):
    benzene = os.path.join(prefix, "test_files", "benzene.xyz")
    catalyst = os.path.join(prefix, "test_files", "catalysts", "tm_single-lig.xyz")
    ligand = os.path.join(prefix, "test_files", "R-Quinox-tBu3.xyz")

    chiral_ring = os.path.join(prefix, "test_files", "chiral_ring.xyz")
    chiral_mol_1 = os.path.join(prefix, "test_files", "chiral_centers_1.xyz")
    chiral_mol_2 = os.path.join(prefix, "test_files", "chiral_centers_2.xyz")
    chiral_mol_3 = os.path.join(prefix, "test_files", "chiral_centers_3.xyz")
    chiral_mol_4 = os.path.join(prefix, "test_files", "chiral_centers_4.xyz")

    def test_BondsFrom(self):
        # test to see if interpolated geometry is correct
        mol = Geometry(self.benzene)
        
        h1 = mol.find('H')[0]

        para_finder = BondsFrom(h1, 5)
        out = mol.find(para_finder, 'H')
        self.assertTrue(all([atom in mol.find('9') for atom in out]))

        meta_finder = BondsFrom(h1, 4)
        out = mol.find(meta_finder, 'H')
        self.assertTrue(all([atom in mol.find('10,11') for atom in out]))

        ortho_finder = BondsFrom(h1, 3)
        out = mol.find(ortho_finder, 'H')
        self.assertTrue(all([atom in mol.find('8,12') for atom in out]))

    def test_WithinBondsOf(self):
        mol = Geometry(self.benzene)

        h1 = mol.find('H')[0]

        out = mol.find(WithinBondsOf(h1, 2))
        self.assertTrue(all([atom in mol.find('1,2,3') for atom in out]))

    def test_BondedTo(self):
        mol = Geometry(self.benzene)

        c1 = mol.find('C')[0]

        out = mol.find(BondedTo(c1))
        self.assertTrue(all([atom in mol.find('2,6,12') for atom in out]))

    def test_WithinRadiusFromPoint(self):
        mol = Geometry(self.benzene)
        mol.coord_shift(-mol.COM())

        out = mol.find(WithinRadiusFromPoint([0, 0, 0], 1.5))
        self.assertTrue(all([atom in mol.find('C') for atom in out]))

    def test_WithinRadiusFromAtom(self):
        mol = Geometry(self.benzene)
        
        c1 = mol.find('C')[0]

        out = mol.find(WithinRadiusFromAtom(c1, 1.5))
        self.assertTrue(all([atom in mol.find('1,2,6,12') for atom in out]))

    def test_NotAny(self):
        mol = Geometry(self.benzene)

        out = mol.find(NotAny('C', '12'))
        self.assertTrue(all([atom in mol.find('7,8,9,10,11') for atom in out]))

    def test_AnyTransitionMetal(self):
        mol = Geometry(self.catalyst)

        out = mol.find(AnyTransitionMetal())
        self.assertTrue(all([atom in mol.find('Cu') for atom in out]))

    def test_AnyNonTransitionMetal(self):
        mol = Geometry(self.benzene)

        out = mol.find(AnyNonTransitionMetal())
        self.assertTrue(all([atom in mol.find('C,H') for atom in out]))

    def test_VSEPR(self):
        mol = Geometry(self.benzene)

        out = mol.find(VSEPR('trigonal planar'))
        self.assertTrue(all([atom in mol.find('C') for atom in out]))

    def test_BondedElements(self):
        mol = Geometry(self.benzene)

        out = mol.find(BondedElements('C', 'C', 'H'))
        self.assertTrue(all([atom in mol.find('C') for atom in out]))

    def test_NumberOfBonds(self):
        mol = Geometry(self.benzene)

        out = mol.find(NumberOfBonds(1))
        self.assertTrue(all([atom in mol.find('H') for atom in out]))

        out = mol.find(NumberOfBonds(3))
        self.assertTrue(all([atom in mol.find('C') for atom in out]))
    
    def test_ChiralCenters(self):
        mol = Geometry(self.ligand)

        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('1') for atom in out]))


        mol = Geometry(self.chiral_ring)
        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('3,12') for atom in out]))

        #next two tests are finding chiral centers that are chiral
        #b/c of other chiral centers
        mol = Geometry(self.chiral_mol_1)
        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('1,3,5') for atom in out]))

        mol = Geometry(self.chiral_mol_2)
        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('1,4') for atom in out]))

        #chiral centers in rings
        mol = Geometry(self.chiral_mol_3)
        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('17,18') for atom in out]))

        mol = Geometry(self.chiral_mol_4)
        out = mol.find(ChiralCenters())
        self.assertTrue(all([atom in mol.find('9,17,18') for atom in out]))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestFinder("test_BondsFrom"))
    suite.addTest(TestFinder("test_WithinBondsOf"))
    suite.addTest(TestFinder("test_BondedTo"))
    suite.addTest(TestFinder("test_WithinRadiusFromPoint"))
    suite.addTest(TestFinder("test_WithinRadiusFromAtom"))
    suite.addTest(TestFinder("test_NotAny"))
    suite.addTest(TestFinder("test_AnyTransitionMetal"))
    suite.addTest(TestFinder("test_AnyNonTransitionMetal"))
    suite.addTest(TestFinder("test_VSEPR"))
    suite.addTest(TestFinder("test_BondedElements"))
    suite.addTest(TestFinder("test_NumberOfBonds"))
    suite.addTest(TestFinder("test_ChiralCenters"))
    return suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""for constructing headers and footers for input files"""
import re

from AaronTools import addlogger
from AaronTools.const import ELEMENTS, UNIT
from AaronTools.theory import (
    GAUSSIAN_COMMENT,
    GAUSSIAN_CONSTRAINTS,
    GAUSSIAN_COORDINATES,
    GAUSSIAN_GEN_BASIS,
    GAUSSIAN_GEN_ECP,
    GAUSSIAN_POST,
    GAUSSIAN_PRE_ROUTE,
    GAUSSIAN_ROUTE,
    ORCA_BLOCKS,
    ORCA_COMMENT,
    ORCA_COORDINATES,
    ORCA_ROUTE,
    PSI4_AFTER_JOB,
    PSI4_BEFORE_GEOM,
    PSI4_BEFORE_JOB,
    PSI4_COMMENT,
    PSI4_COORDINATES,
    PSI4_JOB,
    PSI4_MOLECULE,
    PSI4_OPTKING,
    PSI4_SETTINGS,
    PSI4_SOLVENT,
    SQM_COMMENT,
    SQM_QMMM,
    QCHEM_MOLECULE,
    QCHEM_REM,
    QCHEM_COMMENT,
    QCHEM_SETTINGS,
    FrequencyJob,
    job_from_string,
)
from AaronTools.utils.utils import combine_dicts

from AaronTools.theory.basis import ECP, BasisSet
from AaronTools.theory.emp_dispersion import EmpiricalDispersion
from AaronTools.theory.grid import IntegrationGrid
from AaronTools.theory.job_types import JobType, SinglePointJob
from AaronTools.theory.method import KNOWN_SEMI_EMPIRICAL, Method, SAPTMethod


@addlogger
class Theory:
    """
    A Theory object can be used to create an input file for different QM software.
    The creation of a Theory object does not depend on the specific QM software
    that is determined when the file is written

    attribute names are the same as initialization keywords (with the exception of ecp, which
    is added to the basis attribute)
    valid initialization keywords are:
    geometry                -   AaronTools Geometry
    charge                  -   total charge
    multiplicity            -   electronic multiplicity
    job_type                -   JobType or list(JobType)

    method                  -   Method object (or str - Method instance will be created)
    basis                   -   BasisSet object (or str - will be set to BasisSet(Basis(keyword)))
    ecp                     -   str parsable by BasisSet.parse_basis_str
    empirical_dispersion    -   EmpiricalDispersion object (or str)
    grid                    -   IntegrationGrid object (or str)
    solvent                 -   ImplicitSolvent object

    memory                  -   int - allocated memory (GB)
    processors              -   int - allocated cores
    """

    ACCEPTED_INIT_KW = [
        "geometry",
        "memory",
        "processors",
        "job_type",
        "solvent",
        "grid",
        "empirical_dispersion",
        "basis",
        "method",
    ]

    # if there's a setting that should be an array and Psi4 errors out
    # if it isn't an array, put it in this list (lower case)
    # generally happens if the setting value isn't a string
    # don't add settings that need > 1 value in the array
    FORCED_PSI4_ARRAY = [
        "cubeprop_orbitals",
        "docc",
        "frac_occ",
    ]

    # commonly used settings that do not take array values
    FORCED_PSI4_SINGLE = [
        "reference",
        "scf_type",
        "freeze_core",
        "diag_method",
        "ex_level",
        "fci",
        "maxiter",
        "t",
        "p",
        "opt_type",
        "dft_radial_points",
        "dft_spherical_points",
    ]
    
    FORCED_PSI4_SOLVENT_SINGLE = [
        "units",
        "codata",
        "type",
        "npzfile",
        "area",
        "scaling",
        "raddiset",
        "minradius",
        "mode",
        "nonequilibrium",
        "solvent",
        "solvertype",
        "matrixsymm",
        "correction",
        "diagonalintegrator",
        "diagonalscaling",
        "proberadius",
    ]
    
    # some blocks need to go after the molecule
    # because they refer to atom indices, so orca needs
    # to read the structure first
    ORCA_BLOCKS_AFTER_MOL = ["eprnmr"]
    
    LOG = None

    def __init__(
        self,
        charge=0,
        multiplicity=1,
        method=None,
        basis=None,
        ecp=None,
        empirical_dispersion=None,
        grid=None,
        **kwargs,
    ):
        if not isinstance(charge, list):
            self.charge = int(charge)
        else:
            self.charge = charge

        if not isinstance(multiplicity, list):
            self.multiplicity = int(multiplicity)
        else:
            self.multiplicity = multiplicity

        self.method = None
        self.basis = None
        self.empirical_dispersion = None
        self.grid = None
        self.solvent = None
        self.job_type = None
        self.processors = None
        self.memory = None
        self.geometry = None
        self.kwargs = {}

        for key in self.ACCEPTED_INIT_KW:
            if key in kwargs:
                self.__setattr__(key, kwargs[key])
                del kwargs[key]
            else:
                self.__setattr__(key, None)

        self.kwargs = kwargs

        if isinstance(self.processors, str):
            processors = re.search(r"(\d+)", self.processors)
            if processors:
                self.processors = processors.group(1)

        if isinstance(self.memory, str):
            memory = re.search(r"(\d+)", self.memory)
            if memory:
                self.memory = memory.group(1)

        # if method, basis, etc aren't the expected classes, make them so
        if method is not None:
            if not isinstance(method, Method):
                self.method = Method(
                    method, method.upper() in KNOWN_SEMI_EMPIRICAL
                )
            else:
                self.method = method

        if grid is not None:
            if not isinstance(grid, IntegrationGrid):
                self.grid = IntegrationGrid(grid)
            else:
                self.grid = grid

        if basis is not None:
            self.basis = BasisSet(basis=basis)

        if ecp is not None:
            if self.basis is None:
                self.basis = BasisSet(ecp=ecp)
            else:
                self.basis.ecp = BasisSet(ecp=ecp).ecp

        if empirical_dispersion is not None:
            if not isinstance(empirical_dispersion, EmpiricalDispersion):
                self.empirical_dispersion = EmpiricalDispersion(
                    empirical_dispersion
                )
            else:
                self.empirical_dispersion = empirical_dispersion

        if self.job_type is not None:
            if isinstance(self.job_type, str):
                self.job_type = job_from_string(self.job_type)
                
            if isinstance(self.job_type, JobType):
                self.job_type = [self.job_type]

            # for i, job1 in enumerate(self.job_type):
            #     for job2 in self.job_type[i + 1 :]:
            #         if type(job1) is type(job2):
            #             self.LOG.warning(
            #                 "multiple jobs of the same type: %s, %s"
            #                 % (str(job1), str(job2))
            #             )

    def __setattr__(self, attr, val):
        if isinstance(val, str):
            if attr == "method":
                super().__setattr__(attr, Method(val))
            elif attr == "basis":
                super().__setattr__(attr, BasisSet(val))
            elif attr == "empirical_dispersion":
                super().__setattr__(attr, EmpiricalDispersion(val))
            elif attr == "grid":
                super().__setattr__(attr, IntegrationGrid(val))
            elif attr == "job_type" and isinstance(val, str):
                super().__setattr__(attr, [job_from_string(val)])
            else:
                super().__setattr__(attr, val)
        elif attr == "job_type" and isinstance(val, JobType):
            super().__setattr__(attr, [val])
        else:
            super().__setattr__(attr, val)

    def __eq__(self, other):
        if self.method != other.method:
            # print("method")
            return False
        if self.charge != other.charge:
            # print("charge")
            return False
        if self.multiplicity != other.multiplicity:
            # print("multiplicity")
            return False
        if self.basis != other.basis:
            # print("basis")
            return False
        if self.empirical_dispersion != other.empirical_dispersion:
            # print("disp")
            print(self.empirical_dispersion, other.empirical_dispersion)
            return False
        if self.grid != other.grid:
            # print("grid")
            return False
        if self.solvent != other.solvent:
            # print("solvent")
            return False
        if self.processors != other.processors:
            # print("procs")
            return False
        if self.memory != other.memory:
            # print("mem")
            return False
        if self.job_type and other.job_type:
            if len(self.job_type) != len(other.job_type):
                return False
            for job1, job2 in zip(self.job_type, other.job_type):
                if job1 != job2:
                    return False

        return True

    def copy(self):
        new_dict = dict()
        new_kwargs = dict()
        for key, value in self.__dict__.items():
            try:
                if key == "kwargs":
                    new_kwargs = value.copy()
                else:
                    new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict, **new_kwargs)
 
    def make_header(
        self,
        geom=None,
        style="gaussian",
        conditional_kwargs=None,
        sanity_check_method=False,
        **kwargs,
    ):
        """
        geom: Geometry
        style: str, gaussian, orca, psi4, or sqm
        conditional_kwargs: dict - keys are ORCA_*, PSI4_*, or GAUSSIAN_*
            items in conditional_kwargs will only be added
            to the input if they would otherwise be preset
            e.g. if self.job_type is FrequencyJob and a Gaussian
            input file is being written,
            conditional_kwargs = {GAUSSIAN_ROUTE:{'opt':['noeigentest']}}
            will not add opt=noeigentest to the route
            but if it's an OptimizationJob, it will add opt=noeigentest
        sanity_check_method: bool, check if method is available in recent version
                             of the target software package (Psi4 checks when its
                             footer is created)
        kwargs: keywords are GAUSSIAN_*, ORCA_*, PSI4_*, or QCHEM_*
        """
        if geom is None:
            geom = self.geometry

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if geom is not None:
            self.geometry = geom
        if self.basis is not None:
            self.basis.refresh_elements(self.geometry)

        kwargs = combine_dicts(self.kwargs, kwargs)

        other_kw_dict = {}
        for keyword in kwargs:
            if (
                keyword.startswith("PSI4_")
                or keyword.startswith("ORCA_")
                or keyword.startswith("GAUSSIAN_")
                or keyword.startswith("QCHEM_")
            ):
                new_kw = eval(keyword)
                other_kw_dict[new_kw] = kwargs[keyword]

            elif hasattr(self, keyword):
                if keyword == "method":
                    self.method = Method(kwargs[keyword])

                elif keyword == "basis":
                    self.basis = BasisSet(kwargs[keyword])

                elif keyword == "grid":
                    self.grid = IntegrationGrid(kwargs[keyword])

                elif keyword == "empirical_dispersion":
                    self.grid = EmpiricalDispersion(kwargs[keyword])

                elif keyword == "job_type":
                    if isinstance(kwargs[keyword], JobType):
                        self.job_type = [kwargs[keyword]]
                    else:
                        self.job_type = kwargs[keyword]

                elif keyword in self.ACCEPTED_INIT_KW:
                    setattr(self, keyword, kwargs[keyword])

            else:
                other_kw_dict[keyword] = kwargs[keyword]

        if style == "gaussian":
            return self.get_gaussian_header(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "orca":
            return self.get_orca_header(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "psi4":
            return self.get_psi4_header(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "sqm":
            return self.get_sqm_header(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "qchem":
            return self.get_qchem_header(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        raise NotImplementedError("no get_header method for style: %s" % style)

    def make_molecule(
        self,
        geom=None,
        style="gaussian",
        conditional_kwargs=None,
        **kwargs,
    ):
        """
        geom: Geometry()
        style: gaussian, psi4, or sqm
        conditional_kwargs: dict() of keyword: value pairs
        kwargs: keywords are GAUSSIAN_*, ORCA_*, PSI4_*, or QCHEM_*
        """
        if geom is None:
            geom = self.geometry

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.basis is not None:
            self.basis.refresh_elements(geom)

        kwargs = combine_dicts(self.kwargs, kwargs)

        other_kw_dict = {}
        for keyword in kwargs:
            if (
                keyword.startswith("PSI4_")
                or keyword.startswith("ORCA_")
                or keyword.startswith("GAUSSIAN_")
                or keyword.startswith("QCHEM_")
            ):
                new_kw = eval(keyword)
                other_kw_dict[new_kw] = kwargs[keyword]

            elif hasattr(self, keyword):
                if keyword == "method":
                    self.method = Method(kwargs[keyword])

                elif keyword == "basis":
                    self.basis = BasisSet(kwargs[keyword])

                elif keyword == "grid":
                    self.grid = IntegrationGrid(kwargs[keyword])

                elif keyword == "empirical_dispersion":
                    self.grid = EmpiricalDispersion(kwargs[keyword])

                elif keyword == "job_type":
                    if isinstance(kwargs[keyword], JobType):
                        self.job_type = [kwargs[keyword]]
                    else:
                        self.job_type = kwargs[keyword]

                elif keyword in self.ACCEPTED_INIT_KW:
                    setattr(self, keyword, kwargs[keyword])

            else:
                other_kw_dict[keyword] = kwargs[keyword]

        if style == "gaussian":
            return self.get_gaussian_molecule(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "psi4":
            return self.get_psi4_molecule(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "sqm":
            return self.get_sqm_molecule(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "qchem":
            return self.get_qchem_molecule(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        NotImplementedError("no get_molecule method for style: %s" % style)

    def make_footer(
        self,
        geom=None,
        style="gaussian",
        conditional_kwargs=None,
        sanity_check_method=False,
        **kwargs,
    ):
        """
        geom: Geometry
        style: str, gaussian or psi4
        conditional_kwargs: dict, see make_header
        sanity_check_method: bool, check if method is available in recent version
                             of the target software package (Psi4 only)
        kwargs: keywords are GAUSSIAN_*, ORCA_*, or PSI4_*
        """
        if geom is None:
            geom = self.geometry

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.basis is not None:
            self.basis.refresh_elements(geom)

        kwargs = combine_dicts(self.kwargs, kwargs)

        other_kw_dict = {}
        for keyword in kwargs:
            if (
                keyword.startswith("PSI4_")
                or keyword.startswith("ORCA_")
                or keyword.startswith("GAUSSIAN_")
            ):
                new_kw = eval(keyword)
                other_kw_dict[new_kw] = kwargs[keyword]

            elif hasattr(self, keyword):
                if keyword == "method":
                    self.method = Method(kwargs[keyword])

                elif keyword == "basis":
                    self.basis = BasisSet(kwargs[keyword])

                elif keyword == "grid":
                    self.grid = IntegrationGrid(kwargs[keyword])

                elif keyword == "empirical_dispersion":
                    self.grid = EmpiricalDispersion(kwargs[keyword])

                elif keyword == "job_type":
                    if isinstance(kwargs[keyword], JobType):
                        self.job_type = [kwargs[keyword]]
                    else:
                        self.job_type = kwargs[keyword]

                elif keyword in self.ACCEPTED_INIT_KW:
                    setattr(self, keyword, kwargs[keyword])

            else:
                other_kw_dict[keyword] = kwargs[keyword]

        if style == "gaussian":
            return self.get_gaussian_footer(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "psi4":
            return self.get_psi4_footer(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        elif style == "orca":
            return self.get_orca_footer(
                conditional_kwargs=conditional_kwargs, **other_kw_dict
            )

        NotImplementedError("no get_footer method for style: %s" % style)

    def get_gaussian_header(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        write Gaussian09/16 input file header (up to charge mult)
        other_kw_dict is a dictionary with file positions (using GAUSSIAN_*)
        corresponding to options/keywords
        returns warnings if a certain feature is not available in Gaussian
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_gaussian()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        if (
            GAUSSIAN_COMMENT not in other_kw_dict
            or not other_kw_dict[GAUSSIAN_COMMENT]
        ):
            if self.geometry.comment:
                other_kw_dict[GAUSSIAN_COMMENT] = [self.geometry.comment]
            else:
                other_kw_dict[GAUSSIAN_COMMENT] = [self.geometry.name]

        # add EmpiricalDispersion info
        if self.empirical_dispersion is not None:
            disp, warning = self.empirical_dispersion.get_gaussian()
            other_kw_dict = combine_dicts(other_kw_dict, disp)
            if warning is not None:
                warnings.append(warning)

        # add Integral(grid=X)
        if self.grid is not None:
            grid, warning = self.grid.get_gaussian()
            other_kw_dict = combine_dicts(other_kw_dict, grid)
            if warning is not None:
                warnings.append(warning)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_gaussian()
            if warning is not None:
                warnings.extend(warning)
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        out_str = ""

        # processors, memory, and other link 0 stuff
        if self.processors is not None:
            out_str += "%%NProcShared=%i\n" % self.processors

        if self.memory:
            out_str += "%%Mem=%iGB\n" % self.memory

        if GAUSSIAN_PRE_ROUTE in other_kw_dict:
            for key in other_kw_dict[GAUSSIAN_PRE_ROUTE]:
                out_str += "%%%s" % key
                if other_kw_dict[GAUSSIAN_PRE_ROUTE][key]:
                    out_str += "=%s" % ",".join(
                        other_kw_dict[GAUSSIAN_PRE_ROUTE][key]
                    )

                if not out_str.endswith("\n"):
                    out_str += "\n"

        # start route line with method
        out_str += "#n "
        if self.method is not None:
            func, warning = self.method.get_gaussian()
            if warning is not None:
                warnings.append(warning)
            warning = self.method.sanity_check_method(func, "gaussian")
            if warning:
                warnings.append(warning)
            out_str += "%s" % func
            if not self.method.is_semiempirical and self.basis is not None:
                (
                    basis_info,
                    basis_warnings,
                ) = self.basis.get_gaussian_basis_info()
                warnings.extend(basis_warnings)
                # check basis elements to make sure no element is
                # in two basis sets or left out of any
                if self.geometry is not None:
                    basis_warning = self.basis.check_for_elements(
                        self.geometry
                    )
                    if basis_warning is not None:
                        warnings.append(basis_warning)

                if GAUSSIAN_ROUTE in basis_info:
                    out_str += "%s" % basis_info[GAUSSIAN_ROUTE]

            out_str += " "

        # add other route options
        # only one option can be specfied
        # e.g. for {'Integral':['grid=X', 'grid=Y']}, only grid=X will be used
        if GAUSSIAN_ROUTE in other_kw_dict.keys():
            for option in other_kw_dict[GAUSSIAN_ROUTE].keys():
                known_opts = []
                out_str += option
                if option.lower() == "opt":
                    # need to specified CalcFC for gaussian ts optimization
                    if any(
                        x.lower() == "ts"
                        for x in other_kw_dict[GAUSSIAN_ROUTE][option]
                    ) and not any(
                        x.lower() == y
                        for y in [
                            "calcfc",
                            "readfc",
                            "rcfc",
                            "readcartesianfc",
                            "calcall",
                            "calchffc",
                        ]
                        for x in other_kw_dict[GAUSSIAN_ROUTE][option]
                    ):
                        other_kw_dict[GAUSSIAN_ROUTE][option].append("CalcFC")

                if other_kw_dict[GAUSSIAN_ROUTE][option] or (
                    other_kw_dict[GAUSSIAN_ROUTE][option]
                    and len(other_kw_dict[GAUSSIAN_ROUTE][option]) == 1
                    and (
                        "=" in other_kw_dict[GAUSSIAN_ROUTE][option][0]
                        or "(" in other_kw_dict[GAUSSIAN_ROUTE][option][0]
                    )
                ):
                    out_str += "=("
                    for x in other_kw_dict[GAUSSIAN_ROUTE][option]:
                        opt = x.split("=")[0]
                        if opt not in known_opts:
                            if known_opts:
                                out_str += ","
                            known_opts.append(opt)
                            out_str += x
                    out_str += ")"

                elif (
                    other_kw_dict[GAUSSIAN_ROUTE][option]
                    and len(other_kw_dict[GAUSSIAN_ROUTE][option]) == 1
                ):
                    out_str += "=%s" % other_kw_dict[GAUSSIAN_ROUTE][option][0]

                out_str += " "

        out_str += "\n\n"

        # add comment, removing any trailing newlines
        if GAUSSIAN_COMMENT in other_kw_dict:
            if other_kw_dict[GAUSSIAN_COMMENT]:
                out_str += "\n".join(
                    [x.rstrip() for x in other_kw_dict[GAUSSIAN_COMMENT]]
                )
            else:
                out_str += "comment"

            if not out_str.endswith("\n"):
                out_str += "\n"

        out_str += "\n"

        # charge mult
        out_str += "%i %i\n" % (int(self.charge), int(self.multiplicity))

        if return_warnings:
            return out_str, warnings
        return out_str

    def get_gaussian_molecule(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        get molecule specification for gaussian input files
        """
        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_gaussian()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        if GAUSSIAN_COORDINATES not in other_kw_dict:
            other_kw_dict[GAUSSIAN_COORDINATES] = {}

        if "coords" not in other_kw_dict[GAUSSIAN_COORDINATES]:
            other_kw_dict[GAUSSIAN_COORDINATES][
                "coords"
            ] = self.geometry.coords

        s = ""

        # atom specs need flag column before coords if any atoms frozen
        has_frozen = False
        for atom in self.geometry.atoms:
            if atom.flag:
                has_frozen = True
                break
        for atom, coord in zip(
            self.geometry.atoms, other_kw_dict[GAUSSIAN_COORDINATES]["coords"]
        ):
            s += "%-2s" % atom.element
            if has_frozen:
                s += " % 2d" % (-1 if atom.flag else 0)
            for val in coord:
                s += "  "
                if isinstance(val, float):
                    s += " %9.5f" % val
                elif isinstance(val, str):
                    s += " %5s" % val
                elif isinstance(val, int):
                    s += " %3i" % val
                else:
                    warnings.append("unknown coordinate type: %s" % type(val))
            s += "\n"

        if (
            "variables" in other_kw_dict[GAUSSIAN_COORDINATES]
            and other_kw_dict[GAUSSIAN_COORDINATES]["variables"]
        ):
            s += "Variable:\n"
            for var in other_kw_dict[GAUSSIAN_COORDINATES]["variables"]:
                s += "%4s = %9.5f\n" % tuple(var)

        if (
            "constants" in other_kw_dict[GAUSSIAN_COORDINATES]
            and other_kw_dict[GAUSSIAN_COORDINATES]["constants"]
        ):
            s += "Constant:\n"
            for var in other_kw_dict[GAUSSIAN_COORDINATES]["constants"]:
                s += "%4s = %9.5f\n" % tuple(var)

        if return_warnings:
            return s, warnings

        return s

    def get_gaussian_footer(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """write footer of gaussian input file"""

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_gaussian()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        out_str = ""
        warnings = []

        # if method is not semi emperical, basis set might be gen or genecp
        # get basis info (will be written after constraints)
        if (
            self.method is not None
            and not self.method.is_semiempirical
            and self.basis is not None
        ):
            basis_info, warnings = self.basis.get_gaussian_basis_info()

        elif (
            self.method is not None
            and not self.method.is_semiempirical
            and self.basis is None
        ):
            basis_info = {}
            warnings.append("no basis specfied")

        out_str += "\n"

        # bond, angle, and torsion constraints
        if GAUSSIAN_CONSTRAINTS in other_kw_dict:
            for constraint in other_kw_dict[GAUSSIAN_CONSTRAINTS]:
                out_str += constraint
                out_str += "\n"

            out_str += "\n"

        # write gen info
        if self.method is not None and not self.method.is_semiempirical:
            if GAUSSIAN_GEN_BASIS in basis_info:
                out_str += basis_info[GAUSSIAN_GEN_BASIS]

                out_str += "\n"

            if GAUSSIAN_GEN_ECP in basis_info:
                out_str += basis_info[GAUSSIAN_GEN_ECP]

                out_str += "\n"

        # post info e.g. for NBOREAD
        if GAUSSIAN_POST in other_kw_dict:
            for item in other_kw_dict[GAUSSIAN_POST]:
                out_str += item
                out_str += " "

            out_str += "\n"

        out_str = out_str.rstrip()

        # new lines
        out_str += "\n\n\n"

        if return_warnings:
            return out_str, warnings
        return out_str

    def get_orca_header(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        get ORCA input file header
        other_kw_dict is a dictionary with file positions (using ORCA_*)
        corresponding to options/keywords
        returns file content and warnings e.g. if a certain feature is not available in ORCA
        returns str of header content
        if return_warnings, returns str, list(warning)
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_orca()
                other_kw_dict = combine_dicts(other_kw_dict, job_dict)

        warnings = []

        # if method isn't semi-empirical, get basis info to write later
        if not self.method.is_semiempirical and self.basis is not None:
            basis_info, basis_warnings = self.basis.get_orca_basis_info()
            warnings.extend(basis_warnings)
            if self.geometry is not None:
                warning = self.basis.check_for_elements(self.geometry)
                if warning is not None:
                    warnings.append(warning)

        else:
            basis_info = {}

        other_kw_dict = combine_dicts(other_kw_dict, basis_info)

        # get grid info
        if self.grid is not None:
            grid_info, warning = self.grid.get_orca()
            if warning is not None:
                warnings.append(warning)

            if any(
                "finalgrid" in x.lower() for x in other_kw_dict[ORCA_ROUTE]
            ):
                grid_info[ORCA_ROUTE].pop(1)

            other_kw_dict = combine_dicts(other_kw_dict, grid_info)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_orca()
            warnings.extend(warning)
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        # dispersion
        if self.empirical_dispersion is not None:
            dispersion, warning = self.empirical_dispersion.get_orca()
            if warning is not None:
                warnings.append(warning)

            other_kw_dict = combine_dicts(other_kw_dict, dispersion)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        # start building input file header
        out_str = ""

        # comment
        if ORCA_COMMENT not in other_kw_dict:
            if self.geometry.comment:
                other_kw_dict[ORCA_COMMENT] = [self.geometry.comment]
            else:
                other_kw_dict[ORCA_COMMENT] = [self.geometry.name]
        for comment in other_kw_dict[ORCA_COMMENT]:
            for line in comment.split("\n"):
                out_str += "#%s\n" % line

        out_str += "! "
        # method
        if self.method is not None:
            func, warning = self.method.get_orca()
            if warning is not None:
                warnings.append(warning)
            warning = self.method.sanity_check_method(func, "orca")
            if warning:
                warnings.append(warning)
            out_str += "%s" % func

        # add other route options
        if ORCA_ROUTE in other_kw_dict:
            used_keywords = []
            for kw in other_kw_dict[ORCA_ROUTE]:
                if any(kw.lower() == used_kw for used_kw in used_keywords):
                    continue
                used_keywords.append(kw.lower())
                out_str += " %s" % kw

        out_str += "\n"

        # procs
        if self.processors is not None:
            out_str += "%%pal\n    nprocs %i\nend\n" % self.processors

            # orca memory is per core, so only specify it if processors are specified
            if self.memory:
                out_str += "%%MaxCore %i\n" % (
                    int(1000 * self.memory / self.processors)
                )

        # add other blocks
        if ORCA_BLOCKS in other_kw_dict:
            for keyword in other_kw_dict[ORCA_BLOCKS]:
                if any(keyword.lower() == name for name in self.ORCA_BLOCKS_AFTER_MOL):
                    continue
                if any(other_kw_dict[ORCA_BLOCKS][keyword]):
                    used_settings = []
                    if keyword == "base":
                        out_str += "%%%s " % keyword
                        if isinstance(
                            other_kw_dict[ORCA_BLOCKS][keyword], str
                        ):
                            out_str += (
                                '"%s"\n' % other_kw_dict[ORCA_BLOCKS][keyword]
                            )
                        else:
                            out_str += (
                                '"%s"\n'
                                % other_kw_dict[ORCA_BLOCKS][keyword][0]
                            )
                    else:
                        out_str += "%%%s\n" % keyword
                        for opt in other_kw_dict[ORCA_BLOCKS][keyword]:
                            if any(
                                keyword.lower() == block_name for block_name in [
                                    "freq", "geom",
                                ]
                            ) and any(
                                opt.split()[0].lower() == prev_opt for prev_opt in used_settings
                            ):
                                continue
                            used_settings.append(opt.split()[0].lower())
                            out_str += "    %s\n" % opt
                        out_str += "end\n"

            out_str += "\n"

        # start of coordinate section - end of header
        out_str += "*xyz %i %i\n" % (self.charge, self.multiplicity)

        if return_warnings:
            return out_str, warnings
        return out_str

    def get_orca_footer(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        get ORCA input file header
        other_kw_dict is a dictionary with file positions (using ORCA_*)
        corresponding to options/keywords
        returns file content and warnings e.g. if a certain feature is not available in ORCA
        returns str of header content
        if return_warnings, returns str, list(warning)
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_orca()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        warnings = []

        # if method isn't semi-empirical, get basis info to write later
        if not self.method.is_semiempirical and self.basis is not None:
            basis_info, basis_warnings = self.basis.get_orca_basis_info()
            warnings.extend(basis_warnings)
            if self.geometry is not None:
                warning = self.basis.check_for_elements(self.geometry)
                if warning is not None:
                    warnings.append(warning)

        else:
            basis_info = {}

        other_kw_dict = combine_dicts(basis_info, other_kw_dict)

        # get grid info
        if self.grid is not None:
            grid_info, warning = self.grid.get_orca()
            if warning is not None:
                warnings.append(warning)

            if any(
                "finalgrid" in x.lower() for x in other_kw_dict[ORCA_ROUTE]
            ):
                grid_info[ORCA_ROUTE].pop(1)

            other_kw_dict = combine_dicts(grid_info, other_kw_dict)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_orca()
            warnings.extend(warning)
            other_kw_dict = combine_dicts(solvent_info, other_kw_dict)

        # dispersion
        if self.empirical_dispersion is not None:
            dispersion, warning = self.empirical_dispersion.get_orca()
            if warning is not None:
                warnings.append(warning)

            other_kw_dict = combine_dicts(dispersion, other_kw_dict)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        # start building input file header
        out_str = "\n"

        # add other blocks
        if ORCA_BLOCKS in other_kw_dict:
            for keyword in other_kw_dict[ORCA_BLOCKS]:
                if not any(keyword.lower() == name for name in self.ORCA_BLOCKS_AFTER_MOL):
                    continue
                if any(other_kw_dict[ORCA_BLOCKS][keyword]):
                    if keyword == "base":
                        out_str += "%%%s " % keyword
                        if isinstance(
                            other_kw_dict[ORCA_BLOCKS][keyword], str
                        ):
                            out_str += (
                                '"%s"\n' % other_kw_dict[ORCA_BLOCKS][keyword]
                            )
                        else:
                            out_str += (
                                '"%s"\n'
                                % other_kw_dict[ORCA_BLOCKS][keyword][0]
                            )
                    else:
                        out_str += "%%%s\n" % keyword
                        for opt in other_kw_dict[ORCA_BLOCKS][keyword]:
                            out_str += "    %s\n" % opt
                        out_str += "end\n"

            out_str += "\n"

        if return_warnings:
            return out_str, warnings
        return out_str

    def get_psi4_header(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        write Psi4 input file
        other_kw_dict is a dictionary with file positions (using PSI4_*)
        corresponding to options/keywords
        returns file content and warnings e.g. if a certain feature is not available in Psi4
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_psi4()
                other_kw_dict = combine_dicts(other_kw_dict, job_dict)

        warnings = []

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_psi4()
            warnings.extend(warning)
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        # get basis info if method is not semi empirical
        if not self.method.is_semiempirical and self.basis is not None:
            basis_info, basis_warnings = self.basis.get_psi4_basis_info(
                isinstance(self.method, SAPTMethod)
            )
            warnings.extend(basis_warnings)
            if self.geometry is not None:
                warning = self.basis.check_for_elements(self.geometry)
                if warning is not None:
                    warnings.append(warning)

            # aux basis sets might have a '%s' b/c the keyword to apply them depends on
            # the method - replace %s with the appropriate thing for the method
            for key in basis_info:
                for i in range(0, len(basis_info[key])):
                    if "%s" in basis_info[key][i]:
                        if "cc" in self.method.name.lower():
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "cc"
                            )

                        elif "dct" in self.method.name.lower():
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "dct"
                            )

                        elif "mp2" in self.method.name.lower():
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "mp2"
                            )

                        elif isinstance(self.method, SAPTMethod):
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "sapt"
                            )

                        elif "scf" in self.method.name.lower():
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "scf"
                            )

                        elif "ci" in self.method.name.lower():
                            basis_info[key][i] = basis_info[key][i].replace(
                                "%s", "mcscf"
                            )

        else:
            basis_info = {}

        combined_dict = combine_dicts(other_kw_dict, basis_info)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        # start building input file header
        out_str = ""

        # comment
        if PSI4_COMMENT not in combined_dict:
            if self.geometry.comment:
                combined_dict[PSI4_COMMENT] = [self.geometry.comment]
            else:
                combined_dict[PSI4_COMMENT] = [self.geometry.name]
        for comment in combined_dict[PSI4_COMMENT]:
            for line in comment.split("\n"):
                out_str += "#%s\n" % line

        # procs
        if self.processors is not None:
            out_str += "set_num_threads(%i)\n" % self.processors

        # mem
        if self.memory:
            out_str += "memory %i GB\n" % self.memory

        # before geometry options e.g. basis {} or building a dft method
        if PSI4_BEFORE_GEOM in combined_dict:
            if combined_dict[PSI4_BEFORE_GEOM]:
                for opt in combined_dict[PSI4_BEFORE_GEOM]:
                    out_str += opt
                    out_str += "\n"
                out_str += "\n"

        if return_warnings:
            return out_str, warnings

        return out_str

    def get_psi4_molecule(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        get molecule specification for psi4 input files
        """
        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        use_bohr = False
        pubchem = False
        use_molecule_array = False
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_psi4()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )
        if PSI4_COORDINATES not in other_kw_dict:
            other_kw_dict[PSI4_COORDINATES] = {}

        if "coords" not in other_kw_dict[PSI4_COORDINATES]:
            other_kw_dict[PSI4_COORDINATES]["coords"] = self.geometry.coords

        s = ""

        if isinstance(self.method, SAPTMethod) and not hasattr(
            self.charge, "__iter__"
        ):
            warnings.append(
                "for a SAPTMethod, charge and multiplicity should both be lists\n"
                "with the first item being the overall charge/multiplicity and\n"
                "subsequent items being the charge/multiplicity of the\n"
                "corresponding monomer"
            )
            return s, warnings

        if (
            isinstance(self.method, SAPTMethod)
            and sum(self.multiplicity[1:]) - len(self.multiplicity[1:]) + 1
            > self.multiplicity[0]
        ):
            use_molecule_array = True
            s += "mol = psi4.core.Molecule.from_arrays(\n"
            s += "    molecular_multiplicity=%i,\n" % self.multiplicity[0]
            s += "    molecular_charge=%i,\n" % self.charge[0]
            if PSI4_MOLECULE in other_kw_dict:
                for keyword in other_kw_dict[PSI4_MOLECULE]:
                    if other_kw_dict[keyword]:
                        s += "    %s=%s,\n" % (
                            keyword,
                            repr(other_kw_dict[keyword][0]),
                        )

        else:
            s += "molecule {\n"
            if isinstance(self.method, SAPTMethod):
                if not hasattr(self.charge, "__iter__"):
                    warnings.append(
                        "for a SAPTMethod, charge and multiplicity should both be lists\n"
                        "with the first item being the overall charge/multiplicity and\n"
                        "subsequent items being the charge/multiplicity of the\n"
                        "corresponding monomer"
                    )
                s += "    %2i %i\n" % (self.charge[0], self.multiplicity[0])
                if len(self.charge) > 1 and self.charge[0] != sum(
                    self.charge[1:]
                ):
                    warnings.append(
                        "total charge is not equal to sum of monomer charges"
                    )
            else:
                s += "    %2i %i\n" % (self.charge, self.multiplicity)

            if PSI4_MOLECULE in other_kw_dict:
                for keyword in other_kw_dict[PSI4_MOLECULE]:
                    if other_kw_dict[PSI4_MOLECULE][keyword]:
                        opt = other_kw_dict[PSI4_MOLECULE][keyword][0]
                        if (
                            "pubchem" in keyword.lower()
                            and not keyword.strip().endswith(":")
                        ):
                            keyword = keyword.strip() + ":"
                            pubchem = True
                        s += "     %s %s\n" % (keyword.strip(), opt)
                        if keyword == "units":
                            if opt.lower() in ["bohr", "au", "a.u."]:
                                use_bohr = True

                    else:
                        s += "     %s\n" % keyword

        if use_molecule_array:
            # psi4 input is VERY different for sapt jobs with the low-spin
            # combination of fragments
            monomers = [comp.atoms for comp in self.geometry.components]
            atoms_in_monomer = []
            seps = []
            for i, m1 in enumerate(self.geometry.components[:-1]):
                seps.append(0)
                for m2 in monomers[: i + 1]:
                    seps[-1] += len(m2)

            s += "    fragment_separators=%s,\n" % repr(seps)
            s += "    elez=%s,\n" % repr(
                [
                    ELEMENTS.index(atom.element)
                    for monomer in monomers
                    for atom in monomer
                ]
            )
            s += "    fragment_multiplicities=%s,\n" % repr(
                self.multiplicity[1:]
            )
            s += "    fragment_charges=%s,\n" % repr(self.charge[1:])
            s += "    geom=["
            i = 0
            for monomer in monomers:
                s += "\n"
                for atom in monomer:
                    if atom not in atoms_in_monomer:
                        atoms_in_monomer.append(atom)
                    else:
                        warnings.append("atom in two monomers: %s" % atom.name)
                    ndx = self.geometry.atoms.index(atom)
                    coord = other_kw_dict[PSI4_COORDINATES]["coords"][ndx]
                    for val in coord:
                        s += "       "
                        if isinstance(val, float):
                            if use_bohr:
                                s += "%9.5f," % (val / UNIT.A0_TO_BOHR)
                            else:
                                s += "%9.5f," % val
                        else:
                            warnings.append(
                                "unknown coordinate type: %s" % type(val)
                            )
                    s += "\n"

            s += "    ],\n"
            s += ")\n\n"
            s += "activate(mol)\n"

            if len(atoms_in_monomer) != len(self.geometry.atoms):
                from AaronTools.finders import NotAny

                warnings.append(
                    "there are atoms not in any monomers: %s"
                    % (
                        ", ".join(
                            [
                                atom.name
                                for atom in self.geometry.find(
                                    NotAny(atoms_in_monomer)
                                )
                            ]
                        )
                    )
                )

        elif isinstance(self.method, SAPTMethod):
            monomers = [comp.atoms for comp in self.geometry.components]
            atoms_in_monomer = []
            for monomer, mult, charge in zip(
                monomers, self.multiplicity[1:], self.charge[1:]
            ):
                s += "    --\n"
                s += "    %2i %i\n" % (charge, mult)
                for atom in monomer:
                    s += "    %2s" % atom.element
                    if atom not in atoms_in_monomer:
                        atoms_in_monomer.append(atom)
                    else:
                        warnings.append("atom in two monomers: %s" % atom.name)
                    ndx = self.geometry.atoms.index(atom)
                    coord = other_kw_dict[PSI4_COORDINATES]["coords"][ndx]
                    for val in coord:
                        s += "     "
                        if isinstance(val, float):
                            if use_bohr:
                                s += " %9.5f" % (val / UNIT.A0_TO_BOHR)
                            else:
                                s += " %9.5f" % val
                        elif isinstance(val, str):
                            s += "    %9s" % val
                        else:
                            warnings.append(
                                "unknown coordinate type: %s" % type(val)
                            )
                    s += "\n"

            if "variables" in other_kw_dict[PSI4_COORDINATES]:
                for (name, val, angstrom) in other_kw_dict[PSI4_COORDINATES][
                    "variables"
                ]:
                    if use_bohr and angstrom:
                        val /= UNIT.A0_TO_BOHR
                    s += "     %3s = %9.5f\n" % (name, val)

            s += "}\n"

            if len(atoms_in_monomer) != len(self.geometry.atoms):
                from AaronTools.finders import NotAny

                warnings.append(
                    "there are atoms not in any monomers: %s"
                    % (
                        ", ".join(
                            [
                                atom.name
                                for atom in self.geometry.find(
                                    NotAny(atoms_in_monomer)
                                )
                            ]
                        )
                    )
                )

        elif not pubchem:
            for atom, coord in zip(
                self.geometry.atoms, other_kw_dict[PSI4_COORDINATES]["coords"]
            ):
                s += "    %2s" % atom.element
                for val in coord:
                    s += "     "
                    if isinstance(val, float):
                        s += " %9.5f" % val
                    elif isinstance(val, str):
                        s += " %9s" % val
                    else:
                        warnings.append(
                            "unknown coordinate type: %s" % type(val)
                        )
                s += "\n"

            if "variables" in other_kw_dict[PSI4_COORDINATES]:
                for (name, val, angstrom) in other_kw_dict[PSI4_COORDINATES][
                    "variables"
                ]:
                    if use_bohr and angstrom:
                        val /= UNIT.A0_TO_BOHR
                    s += "     %3s = %9.5f\n" % (name, val)

            s += "}\n"
        else:
            s += "}\n"

        if return_warnings:
            return s, warnings
        return s

    def get_psi4_footer(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        get psi4 footer
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_psi4()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_psi4()
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        # grid
        if self.grid is not None:
            grid_info, warning = self.grid.get_psi4()
            if warning is not None:
                warnings.append(warning)
            other_kw_dict = combine_dicts(other_kw_dict, grid_info)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        out_str = "\n"

        # settings
        # a setting will only get added if its list has at least
        # one item, but only the first item will be used
        if PSI4_SETTINGS in other_kw_dict and any(
            other_kw_dict[PSI4_SETTINGS][setting]
            for setting in other_kw_dict[PSI4_SETTINGS]
        ):
            out_str += "set {\n"
            for setting in other_kw_dict[PSI4_SETTINGS]:
                if other_kw_dict[PSI4_SETTINGS][setting]:
                    if isinstance(other_kw_dict[PSI4_SETTINGS][setting], str):
                        val = other_kw_dict[PSI4_SETTINGS][setting]
                    else:
                        if len(
                            other_kw_dict[PSI4_SETTINGS][setting]
                        ) == 1 and (
                            not any(
                                array_setting == setting.strip().lower()
                                for array_setting in self.FORCED_PSI4_ARRAY
                            )
                            or any(
                                single_setting == setting.strip().lower()
                                for single_setting in self.FORCED_PSI4_SINGLE
                            )
                        ):
                            val = other_kw_dict[PSI4_SETTINGS][setting][0]
                        else:
                            # array of values
                            val = "["
                            val += ",".join(
                                [
                                    "%s" % v
                                    for v in other_kw_dict[PSI4_SETTINGS][
                                        setting
                                    ]
                                ]
                            )
                            val += "]"

                    out_str += "    %-20s    %s\n" % (setting, val)

            out_str += "}\n\n"


        if PSI4_SOLVENT in other_kw_dict:
            out_str += "pcm = {\n"
            for setting in other_kw_dict[PSI4_SOLVENT]:
                if other_kw_dict[PSI4_SOLVENT][setting]:
                    if isinstance(other_kw_dict[PSI4_SOLVENT][setting], str):
                        val = other_kw_dict[PSI4_SOLVENT][setting]
                        out_str += "    %s = %s\n" % (setting, val)
                    else:
                        if any(
                            single_setting == setting.strip().lower()
                            for single_setting in self.FORCED_PSI4_SOLVENT_SINGLE
                        ):
                            val = other_kw_dict[PSI4_SOLVENT][setting][0]
                            out_str += "    %s = %s\n" % (setting, val)
                        else:
                            # array of values
                            if not out_str.endswith("\n\n") and not out_str.endswith("{\n"):
                                out_str += "\n"
                            out_str += "    %s {\n" % setting
                            for val in other_kw_dict[PSI4_SOLVENT][setting]:
                                out_str += "        %s\n" % val
                            out_str += "    }\n\n"

            out_str += "}\n\n"


        if PSI4_OPTKING in other_kw_dict and any(
            other_kw_dict[PSI4_OPTKING][setting]
            for setting in other_kw_dict[PSI4_OPTKING]
        ):
            out_str += "set optking {\n"
            for setting in other_kw_dict[PSI4_OPTKING]:
                if other_kw_dict[PSI4_OPTKING][setting]:
                    out_str += "    %-20s    %s\n" % (
                        setting,
                        other_kw_dict[PSI4_OPTKING][setting][0],
                    )

            out_str += "}\n\n"

        # method is method name + dispersion if there is dispersion
        method = self.method.get_psi4()[0]
        if self.empirical_dispersion is not None:
            disp = self.empirical_dispersion.get_psi4()[0]
            if "%s" in method:
                method = method % disp
            else:
                method += disp
        elif "%s" in method:
            method = method.replace("%s", "")

        warning = self.method.sanity_check_method(method, "psi4")
        if warning:
            warnings.append(warning)

        # after job stuff - replace METHOD with method
        if PSI4_BEFORE_JOB in other_kw_dict:
            for opt in other_kw_dict[PSI4_BEFORE_JOB]:
                if "$METHOD" in opt:
                    opt = opt.replace("$METHOD", "'%s'" % method)

                out_str += opt
                out_str += "\n"

        # for each job, start with nrg = f('method'
        # unless return_wfn=True, then do nrg, wfn = f('method'
        if PSI4_JOB in other_kw_dict:
            for func in other_kw_dict[PSI4_JOB].keys():
                if any(
                    [
                        "return_wfn" in kwarg
                        and ("True" in kwarg or "on" in kwarg)
                        for kwarg in other_kw_dict[PSI4_JOB][func]
                    ]
                ):
                    out_str += "nrg, wfn = %s('%s'" % (func, method)
                else:
                    out_str += "nrg = %s('%s'" % (func, method)

                known_kw = []
                for keyword in other_kw_dict[PSI4_JOB][func]:
                    key = keyword.split("=")[0].strip()
                    if key not in known_kw:
                        known_kw.append(key)
                        out_str += ", "
                        out_str += keyword.replace("$METHOD", "'%s'" % method)

                out_str += ")\n"

        # after job stuff - replace METHOD with method
        if PSI4_AFTER_JOB in other_kw_dict:
            for opt in other_kw_dict[PSI4_AFTER_JOB]:
                if "$METHOD" in opt:
                    opt = opt.replace("$METHOD", "'%s'" % method)

                out_str += opt
                out_str += "\n"

        if return_warnings:
            return out_str, warnings
        return out_str

    def get_xtb_cmdline(self, config):
        """
        Uses the config and job type to set command line options for xtb and crest jobs

        Returns a dictionary of option=val pairs; val is None when option doesn't take
        an argument. This dict should be parsed by the caller into the command line
        string.
        """
        if len(self.job_type) > 1:
            raise NotImplementedError(
                "Multiple job types not supported for crest/xtb"
            )
        cmdline = {}
        job_type = self.job_type[0]
        style = config["Job"]["exec_type"]
        if style not in ["xtb", "crest"]:
            raise NotImplementedError(
                "Wrong executable type: %s (can only get command line options "
                "for `xtb` or `crest`)" % style
            )

        # pull in stuff set by resolve_error
        if config._args:
            for arg in config._args:
                cmdline[arg] = None
        if config._kwargs:
            for key, val in config._kwargs.items():
                cmdline[key] = val
        # run types
        if "gfn" in config["Theory"]:
            cmdline["--gfn"] = config["Job"]["gfn"]
        if (
            style == "xtb"
            and hasattr(job_type, "transition_state")
            and job_type.transition_state
        ):
            cmdline["--optts"] = None
        elif style == "xtb":
            cmdline["--opt"] = None

        # charge/mult/temp
        if self.charge != 0:
            cmdline["--chrg"] = self.charge
        if self.multiplicity != 1:
            cmdline["--uhf"] = self.multiplicity - 1
        if style == "crest":
            cmdline["--temp"] = config["Theory"].get(
                "temperature", fallback="298"
            )
        else:
            cmdline["--etemp"] = config["Theory"].get(
                "temperature", fallback="298"
            )

        # screening parameters
        if (
            style == "crest"
            and "energy_cutoff" in config["Job"]
            and config["Job"].getfloat("energy_cutoff") > 6
        ):
            cmdline["--ewin"] = config["Job"]["energy_cutoff"]
        if (
            style == "crest"
            and "rmsd_cutoff" in config["Job"]
            and config["Job"].getfloat("rmsd_cutoff") < 0.125
        ):
            cmdline["--rthr"] = config["Job"]["rmsd_cutoff"]

        # solvent stuff
        if (
            "solvent_model" in config["Theory"]
            and config["Theory"]["solvent_model"] == "alpb"
        ):
            solvent = config["Theory"]["solvent"].split()
            if len(solvent) > 1:
                solvent, ref = solvent
            elif len(solvent) == 1:
                solvent, ref = solvent[0], None
            else:
                raise ValueError
            if solvent.lower() not in [
                "acetone",
                "acetonitrile",
                "aniline",
                "benzaldehyde",
                "benzene",
                "ch2cl2",
                "chcl3",
                "cs2",
                "dioxane",
                "dmf",
                "dmso",
                "ether",
                "ethylacetate",
                "furane",
                "hexandecane",
                "hexane",
                "methanol",
                "nitromethane",
                "octanol",
                "woctanol",
                "phenol",
                "toluene",
                "thf",
                "water",
            ]:
                raise ValueError("%s is not a supported solvent" % solvent)
            if ref is not None and ref.lower() not in ["reference", "bar1m"]:
                raise ValueError(
                    "%s Gsolv reference state not supported" % ref
                )
            if style.lower() == "crest" or ref is None:
                cmdline["--alpb"] = "{}".format(solvent)
            else:
                cmdline["--alpb"] = "{} {}".format(solvent, ref)
        elif (
            "solvent_model" in config["Theory"]
            and config["Theory"]["solvent_model"] == "gbsa"
        ):
            solvent = config["Theory"]["solvent"].split()
            if len(solvent) > 1:
                solvent, ref = solvent
            else:
                solvent, ref = solvent[0], None
            if solvent.lower() not in [
                "acetone",
                "acetonitrile",
                "benzene",
                "ch2cl2",
                "chcl3",
                "cs2",
                "dmf",
                "dmso",
                "ether",
                "h2o",
                "methanol",
                "n-hexane",
                "thf",
                "toluene",
            ]:
                gfn = config["Theory"].get("gfn", fallback="2")
                if gfn != "1" and solvent.lower() in ["benzene"]:
                    raise ValueError("%s is not a supported solvent" % solvent)
                elif gfn != "2" and solvent.lower() in ["DMF", "n-hexane"]:
                    raise ValueError("%s is not a supported solvent" % solvent)
                else:
                    raise ValueError("%s is not a supported solvent" % solvent)
            if ref is not None and ref.lower() not in ["reference", "bar1m"]:
                raise ValueError(
                    "%s Gsolv reference state not supported" % ref
                )
            if style.lower() == "crest" or ref is None:
                cmdline["--gbsa"] = "{}".format(solvent)
            else:
                cmdline["--gbsa"] = "{} {}".format(solvent, ref)

        other = config["Job"].get("cmdline", fallback="").split()
        i = 0
        while i < len(other):
            if other[i].startswith("-"):
                key = other[i]
                cmdline[key] = None
            else:
                cmdline[key] = other[i]
            i += 1
        return cmdline

    def get_xcontrol(self, config, ref=None):
        if len(self.job_type) > 1:
            raise NotImplementedError(
                "Multiple job types not supported for crest/xtb"
            )
        job_type = self.job_type[0]
        return job_type.get_xcontrol(config, ref=ref)

    def get_sqm_header(
            self,
            return_warnings=False,
            conditional_kwargs=None,
            **other_kw_dict,
    ):
        """retruns header, warnings_list for sqm job"""
        
        warnings = []
   

        if conditional_kwargs is None:
            conditional_kwargs = {}

        # job stuff
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict, job_warnings = job.get_sqm()
                warnings.extend(job_warnings)
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        s = ""

        # charge and mult
        other_kw_dict = combine_dicts(
            {
                SQM_QMMM: {
                    "qmcharge": [str(self.charge)],
                    "spin": [str(self.multiplicity)],
                }
            },
            other_kw_dict,
        )

        # comment
        if SQM_COMMENT not in other_kw_dict:
            if self.geometry.comment:
                other_kw_dict[SQM_COMMENT] = [self.geometry.comment]
            else:
                other_kw_dict[SQM_COMMENT] = [self.geometry.name]

        for comment in other_kw_dict[SQM_COMMENT]:
            for line in comment.split("\n"):
                s += "%s" % line
            s += "\n"
        
        # method
        if self.method:
            method = self.method.get_sqm()
            warning = self.method.sanity_check_method(method, "sqm")
            if warning:
                warnings.append(warning)
            other_kw_dict = combine_dicts(
                other_kw_dict,
                {SQM_QMMM: {"qm_theory": [method]}}
            )
        
        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        # options
        s += " &qmmm\n"
        for key in other_kw_dict[SQM_QMMM]:
            if not other_kw_dict[SQM_QMMM][key]:
                continue
            s += "   %s=" % key
            option = other_kw_dict[SQM_QMMM][key][0]
            if re.match("-?\d+", option):
                s += "%s,\n" % option
            elif any(option.lower() == b for b in [".true.", ".false."]):
                s += "%s,\n" % option
            else:
                s += "'%s',\n" % option
        s += " /\n"
        
        return s, warnings
    
    def get_sqm_molecule(
        self,
        **kwargs,
    ):
        """returns molecule specification for sqm input"""
        
        warnings = []
        s = ""

        for atom in self.geometry.atoms:
            s += " %2i  %2s  %9.5f  %9.5f  %9.5f\n" % (
                ELEMENTS.index(atom.element),
                atom.element,
                atom.coords[0],
                atom.coords[1],
                atom.coords[2],
            )
        
        return s.rstrip(), warnings

    def get_qchem_header(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):
        """
        write QChem input file header (up to charge mult)
        other_kw_dict is a dictionary with file positions (using QCHEM_*)
        corresponding to options/keywords
        returns warnings if a certain feature is not available in QChem
        """

        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        job_type_count = 0
        if self.job_type is not None:
            for i, job in enumerate(self.job_type[::-1]):
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_qchem()
                if isinstance(job, FrequencyJob) and job.temperature != 298.15:
                    warnings.append(
                        "thermochemistry data in the output file will be for 298.15 K\n"
                        "in spite of the user setting %.2f K\n" % (job.temperature) + \
                        "free energy corrections can be calculated at different\n"
                        "temperatures using AaronTools grabThermo.py script or\n"
                        "SEQCROW's Procress QM Thermochemistry tool"
                    )
                if QCHEM_REM in job_dict and any(key.lower() == "job_type" for key in job_dict[QCHEM_REM]):
                    job_type_count += 1
                if job_type_count > 1:
                    raise NotImplementedError("cannot put multiple JOB_TYPE entries in one Q-Chem header")
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        if (
            QCHEM_COMMENT not in other_kw_dict
            or not other_kw_dict[QCHEM_COMMENT]
        ):
            if self.geometry.comment:
                other_kw_dict[QCHEM_COMMENT] = [self.geometry.comment]
            else:
                other_kw_dict[QCHEM_COMMENT] = [self.geometry.name]

        if QCHEM_SETTINGS in other_kw_dict:
            other_kw_dict = combine_dicts(
                other_kw_dict,
                {QCHEM_SETTINGS: {QCHEM_COMMENT: other_kw_dict[QCHEM_COMMENT]}},
            )
        else:
            other_kw_dict[QCHEM_SETTINGS] = {QCHEM_COMMENT: other_kw_dict[QCHEM_COMMENT]}

        # add memory info
        if self.memory:
            other_kw_dict = combine_dicts(
                other_kw_dict,
                {QCHEM_REM: {"MEM_TOTAL": str(1000 * self.memory)}}
            )

        # add EmpiricalDispersion info
        if self.empirical_dispersion is not None:
            disp, warning = self.empirical_dispersion.get_qchem()
            other_kw_dict = combine_dicts(other_kw_dict, disp)
            if warning is not None:
                warnings.append(warning)

        # add Integral(grid=X)
        if self.grid is not None:
            grid, warning = self.grid.get_qchem()
            other_kw_dict = combine_dicts(other_kw_dict, grid)
            if warning is not None:
                warnings.append(warning)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_qchem()
            if warning is not None:
                warnings.extend(warning)
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        if self.method is not None:
            func, warning = self.method.get_qchem()
            if warning is not None:
                warnings.append(warning)

            warning = self.method.sanity_check_method(func, "qchem")
            if warning:
                warnings.append(warning)
            
            # Q-Chem seems to still require a basis set for HF-3c
            # if not self.method.is_semiempirical and self.basis is not None:
            if self.basis is not None:
                (
                    basis_info,
                    basis_warnings,
                ) = self.basis.get_qchem_basis_info(self.geometry)
                warnings.extend(basis_warnings)
                # check basis elements to make sure no element is
                # in two basis sets or left out of any
                other_kw_dict = combine_dicts(
                    other_kw_dict,
                    basis_info,
                )
                if self.geometry is not None:
                    basis_warning = self.basis.check_for_elements(
                        self.geometry,
                    )
                    if basis_warning is not None:
                        warnings.append(basis_warning)

            other_kw_dict = combine_dicts(
                other_kw_dict, {QCHEM_REM: {"METHOD": func}},
            )

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        out_str = ""

        if QCHEM_REM in other_kw_dict and QCHEM_SETTINGS in other_kw_dict:
            other_kw_dict[QCHEM_SETTINGS] = combine_dicts(
                {"rem": other_kw_dict[QCHEM_REM]}, other_kw_dict[QCHEM_SETTINGS],
            )
        
        elif QCHEM_REM in other_kw_dict:
            other_kw_dict[QCHEM_SETTINGS] = {"rem": other_kw_dict[QCHEM_REM]}
        
        else:
            warnings.append("no REM section")

        if self.memory:
            other_kw_dict = combine_dicts(
                other_kw_dict,
                {"rem": {"MEM_TOTAL": "%i" % (1000 * self.memory)}}
            )

        if QCHEM_SETTINGS in other_kw_dict:
            for section in other_kw_dict[QCHEM_SETTINGS]:
                settings = other_kw_dict[QCHEM_SETTINGS][section]
                out_str += "$%s\n" % section
                for setting in settings:
                    if not setting:
                        continue
                    if isinstance(settings, dict):
                        opt = settings[setting]
                        if not opt:
                            continue
                        if isinstance(opt, str):
                            val = opt
                            out_str += "    %-20s =   %s\n" % (setting, val)
                        elif isinstance(opt, dict):
                            for s, v in opt.items():
                                out_str += "    %-20s =   %s\n" % (s, v)
                        else:
                            if len(opt) == 1:
                                val = opt[0]
                                out_str += "    %-20s =   %s\n" % (setting, val)
                            elif not opt:
                                out_str += "    %-20s\n" % setting
                            else:
                                if section.lower() == "rem" and setting.lower() == "job_type":
                                    raise NotImplementedError(
                                        "cannot put multiple JOB_TYPE entries in one Q-Chem header"
                                    )
                                out_str += "    %-20s =   %s\n" % (setting, ", ".join(opt))

    
                    elif hasattr(setting, "__iter__") and not isinstance(setting, str):
                        for val in setting:
                            out_str += "    %s\n" % val
                    
                    else:
                        out_str += "    %s\n" % setting
    
                out_str += "$end\n\n"
        
        return out_str, warnings
    
    def get_qchem_molecule(
        self,
        return_warnings=False,
        conditional_kwargs=None,
        **other_kw_dict,
    ):

        if conditional_kwargs is None:
            conditional_kwargs = {}

        warnings = []
        if self.job_type is not None:
            for job in self.job_type[::-1]:
                if hasattr(job, "geometry"):
                    job.geometry = self.geometry

                job_dict = job.get_qchem()
                other_kw_dict = combine_dicts(job_dict, other_kw_dict)

        if (
            GAUSSIAN_COMMENT not in other_kw_dict
            or not other_kw_dict[GAUSSIAN_COMMENT]
        ):
            if self.geometry.comment:
                other_kw_dict[GAUSSIAN_COMMENT] = [self.geometry.comment]
            else:
                other_kw_dict[GAUSSIAN_COMMENT] = [self.geometry.name]

        # add EmpiricalDispersion info
        if self.empirical_dispersion is not None:
            disp, warning = self.empirical_dispersion.get_qchem()
            other_kw_dict = combine_dicts(other_kw_dict, disp)
            if warning is not None:
                warnings.append(warning)

        # add Integral(grid=X)
        if self.grid is not None:
            grid, warning = self.grid.get_qchem()
            other_kw_dict = combine_dicts(other_kw_dict, grid)
            if warning is not None:
                warnings.append(warning)

        # add implicit solvent
        if self.solvent is not None:
            solvent_info, warning = self.solvent.get_qchem()
            if warning is not None:
                warnings.extend(warning)
            other_kw_dict = combine_dicts(other_kw_dict, solvent_info)

        other_kw_dict = combine_dicts(
            other_kw_dict, conditional_kwargs, dict2_conditional=True
        )

        out_str = "$molecule\n    %i %i\n" % (
            self.charge, self.multiplicity
        )

        if QCHEM_MOLECULE in other_kw_dict:
            for line in other_kw_dict[QCHEM_MOLECULE]:
                out_str += "    %-20s\n" % line
        elif not self.geometry:
            warnings.append("no molecule")
        
        for atom in self.geometry.atoms:
            out_str += "    %-2s" % atom.element
            out_str += "   %9.5f    %9.5f    %9.5f\n" % tuple(atom.coords)

        out_str += "$end\n"
        
        return out_str, warnings

    
    



#!/usr/bin/env python3
import os
import unittest

import numpy as np
from AaronTools.fileIO import FileReader
from AaronTools.geometry import Geometry
from AaronTools.test import TestWithTimer, prefix


class TestOrbitals(TestWithTimer):
    small_mol = os.path.join(prefix, "test_files", "formaldehyde.fchk")

    def test_fchk_mo(self):
        ref = np.array([
            [ 1.98430354e-03, 1.14245330e+01, -2.01456705e-04, -2.01441878e-04],
            [ 7.31370333e+00, -1.94029767e-03, -1.36633289e-03, -1.36633695e-03],
            [ 0.9581908, 2.56044193, -0.03363992, -0.03363798],
            [ 1.33606002, -1.13133406, -0.24243181, -0.24241961],
            [ 2.36431044e-05, -2.45960625e-05, 2.49700776e-01, -2.49702951e-01],
            [ 0.26057303, -1.10826257, 0.10474883, 0.10475366],
            [ 2.41646200e-05, 1.99531504e-05, -2.00620098e-06, -2.00371274e-06],
            [-2.08098291e-06, 4.79409908e-07, 2.33169314e-01, -2.33170704e-01],
            [ 6.12950700e-05, -1.35701471e-05, 1.58944944e-05, 1.58759698e-05],
            [ 0.88931739, -0.05547194, 0.08981646, 0.08982103],
        ])
        
        fr = FileReader(TestOrbitals.small_mol, just_geom=False)
        geom = Geometry(fr)
        orbits = fr.other["orbitals"]
        coords = geom.coords
        for mo in range(0, 10):
            vals = orbits.mo_value(mo, coords)
            self.assertTrue(np.linalg.norm(vals - ref[mo]) < 1e-6)

    def test_fchk_density(self):
        ref = np.array([112.52268929, 279.16796565, 0.37519492, 0.37518832])
        fr = FileReader(TestOrbitals.small_mol, just_geom=False)
        geom = Geometry(fr)
        orbits = fr.other["orbitals"]
        coords = geom.coords
        vals = orbits.density_value(coords)
        self.assertTrue(np.linalg.norm(vals - ref) < 1e-6)

    def test_fchk_fukui_dual(self):
        ref = np.array([ 0.03395042, -0.00139618, -0.10751031, -0.10751156])
        fr = FileReader(TestOrbitals.small_mol, just_geom=False)
        geom = Geometry(fr)
        orbits = fr.other["orbitals"]
        coords = geom.coords
        vals = orbits.fukui_dual_value(coords)
        self.assertTrue(np.linalg.norm(vals - ref) < 1e-6)


if __name__ == "__main__":
    unittest.main()
